#!/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

import yaml


class _ConfigSingleton(type):
    """
    Config singleton metaclass
    """
    _instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(_ConfigSingleton, cls).__call__(*args, **kwargs)
        return cls._instance

    @classmethod
    def set_config(cls, config):
        """
        Set config which should be used in within Config singleton

        :param config: configuration that should be used
        :type: dict
        """
        # set _config before the singleton is instantiated
        assert(cls._instance is None)
        cls._config = config


class Config(metaclass=_ConfigSingleton):
    """
    Configuration supplied by user
    """
    _config = None

    def __init__(self):
        self._raw_config = None

        if self._config is None:
            return

        with open(self._config) as f:
            self._raw_config = yaml.load(f, Loader=yaml.SafeLoader)

    def style_task(self):
        """
        Return style for tasks in the graph, see graphviz styling options

        :return: style definition
        :rtype: dict
        """
        if self._raw_config is None or 'style' not in self._raw_config:
            return {}
        ret = self._raw_config['style'].get('task', {})
        return ret if ret is not None else {}

    def style_flow(self):
        """
        Return style for a flow node in the graph, see graphviz styling options

        :return: style definition
        :rtype: dict
        """
        default = {'style': 'filled', 'fillcolor': '#CCCCCC'}
        if self._raw_config is None or 'style' not in self._raw_config:
            return default
        ret = self._raw_config['style'].get('flow', default)
        return ret if ret is not None else default

    def style_condition(self):
        """
        Return style for conditions in the graph, see graphviz styling options

        :return: style definition
        :rtype: dict
        """
        default_style = {'shape': 'hexagon'}
        if self._raw_config is None or 'style' not in self._raw_config:
            return default_style
        ret = self._raw_config['style'].get('condition', default_style)
        return ret if ret is not None else default_style

    def style_storage(self):
        """
        Return style for storage in the graph, see graphviz styling options

        :return: style definition
        :rtype: dict
        """
        default_style = {'shape': 'folder', 'color': '#AAAAAA'}
        if self._raw_config is None or 'style' not in self._raw_config:
            return default_style
        ret = self._raw_config['style'].get('storage', default_style)
        return ret if ret is not None else {}

    def style_edge(self):
        """
        Return style for edges in the graph, see graphviz styling options

        :return: style definition
        :rtype: dict
        """
        if self._raw_config is None or 'style' not in self._raw_config:
            return {}
        ret = self._raw_config['style'].get('edge', {})
        return ret if ret is not None else {}

    def style_store_edge(self):
        """
        Return style for edges that lead to a storage in the graph, see graphviz styling options

        :return: style definition
        :rtype: dict
        """
        default_style = {'color': '#AAAAAA', 'style': 'dashed'}
        if self._raw_config is None or 'style' not in self._raw_config:
            return default_style
        ret = self._raw_config['style'].get('store-edge', default_style)
        return ret if ret is not None else default_style

    def style_graph(self):
        """
        Return style for the whole graph, see graphviz styling options

        :return: style definition
        :rtype: dict
        """
        if self._raw_config is None or 'style' not in self._raw_config:
            return {}
        ret = self._raw_config['style'].get('graph', {})
        return ret if ret is not None else {}

    def style_fallback_edge(self):
        """
        Return style for fallback edges

        :return: style definition
        :rtype: dict
        """
        default_style = {'color': '#FF0000', 'style': 'dashed'}
        if self._raw_config is None or 'style' not in self._raw_config:
            return default_style
        ret = self._raw_config['style'].get('fallback-edge', default_style)
        return ret if ret is not None else default_style

    def style_fallback_node(self):
        """
        Return style for fallback node

        :return: style definition
        :rtype: dict
        """
        default_style = {'color': '#FF0000', 'shape': 'point'}
        if self._raw_config is None or 'style' not in self._raw_config:
            return default_style
        ret = self._raw_config['style'].get('fallback-edge', default_style)
        return ret if ret is not None else default_style

    def style_fallback_true(self):
        """
        Return style for fallback true node

        :return: style definition
        :rtype: dict
        """
        default_style = {'color': '#222222', 'shape': 'doubleoctagon'}
        if self._raw_config is None or 'style' not in self._raw_config:
            return default_style
        ret = self._raw_config['style'].get('fallback-edge', default_style)
        return ret if ret is not None else default_style



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

import importlib
from .helpers import get_function_arguments

_EXPECTED_STRATEGY_FUNC_ARGS = {'previous_retry', 'active_nodes', 'failed_nodes',
                                'new_started_nodes', 'new_fallback_nodes'}

# Celery's default queue
_DEFAULT_CELERY_QUEUE = 'celery'


class GlobalConfig(object):
    """
    User global configuration stated in YAML file
    """
    predicates_module = 'selinonlib.predicates'

    strategy_module = 'selinonlib.strategies'
    strategy_function = 'biexponential_increase'
    strategy_func_args = {'start_retry': 2, 'max_retry': 120}

    default_task_queue = _DEFAULT_CELERY_QUEUE
    config_dispatcher_queue = _DEFAULT_CELERY_QUEUE

    _trace_logging = None
    _trace_import = None
    _trace_function = None
    _trace_storage = None

    def __init__(self):
        raise NotImplementedError("Cannot instantiate global config")

    @classmethod
    def _parse_strategy(cls, strategy_dict):
        """
        Parse strategy entry

        :param strategy_dict: strategy entry in config to be parsed
        """
        if strategy_dict is None:
            return

        if not strategy_dict:
            raise ValueError('Strategy not defined properly in global configuration section')

        if 'name' not in strategy_dict:
            raise ValueError('Sampling strategy stated in global configuration but no strategy name defined')

        if 'import' in strategy_dict:
            cls.strategy_module = strategy_dict['import']

        module = importlib.import_module(cls.strategy_module)

        strategy_func = getattr(module, strategy_dict['name'])

        if strategy_dict.get('args') is not None and not isinstance(strategy_dict['args'], dict):
            raise ValueError('Arguments to strategy function should be stated as dict')
        user_args = strategy_dict.get('args', {})

        user_args_keys = user_args.keys()
        func_args = set(get_function_arguments(strategy_func))

        if (func_args - user_args_keys) != _EXPECTED_STRATEGY_FUNC_ARGS:
            raise ValueError('Unknown or invalid arguments supplied to sampling strategy function, expected %s, got %s'
                             % ((func_args - _EXPECTED_STRATEGY_FUNC_ARGS), set(user_args_keys)))
        else:
            cls.strategy_func_args = user_args

    @classmethod
    def dump_trace(cls, output, config_name, indent_count=0):
        """
        Dump trace configuration to output stream

        :param output: output stream to write to
        :param config_name: name of configuration class instance to be referenced when initializing trace
        :param indent_count: indentation that should be used to indent source
        """
        indent = indent_count * 4 * " "

        if cls._trace_logging:
            output.write('%s%s.trace_by_logging()\n' % (indent, config_name))
            return

        if cls._trace_storage:
            output.write('%s%s.trace_by_func(functools.partial(%s.%s, %s)\n'
                         % (indent, config_name, cls._trace_storage.class_name,
                            cls._trace_function, cls._trace_storage.var_name))

        if cls._trace_import:
            output.write('%sfrom %s import %s\n' % (indent, cls._trace_import, cls._trace_function))
            output.write('%s%s.trace_by_func(%s)\n' % (indent, config_name, cls._trace_function))

    @classmethod
    def _parse_trace(cls, system, trace_record):
        """
        Parse trace configuration entry

        @param system: system instance for which the parsing is done (for storage lookup)
        @param trace_record: trace record to be parsed
        """
        if trace_record is None:
            raise ValueError('Trace not defined properly in global configuration section')

        if trace_record is False:
            return

        if trace_record is True:
            cls._trace_logging = True
            return

        if 'storage' in trace_record:
            if 'method' in trace_record:
                cls._trace_function = trace_record['method']
            else:
                cls._trace_function = 'trace'

            storage = system.storage_by_name(trace_record['storage'])
            cls._trace_storage = storage
        else:
            if 'import' not in trace_record:
                raise ValueError('Expected import definition if trace is not logging nor storage')

            if 'function' not in trace_record:
                raise ValueError('Expected function definition if trace is not logging nor storage')

            cls._trace_import = trace_record['import']
            cls._trace_function = trace_record['function']

    @classmethod
    def from_dict(cls, system, d):
        """
        Parse global configuration from a dictionary

        :param system: system instance for storage lookup
        :param d: dictionary containing global configuration as stated in YAML config file
        """
        if 'predicates_module' in d:
            cls.predicates_module = d['predicates_module']

        if 'strategy' in d:
            cls._parse_strategy(d['strategy'])

        if 'trace' in d:
            cls._parse_trace(system, d['trace'])

        cls.dispatcher_queue = d.get('dispatcher_queue', _DEFAULT_CELERY_QUEUE)
        cls.default_task_queue = d.get('default_task_queue', _DEFAULT_CELERY_QUEUE)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################


import abc
from .helpers import dict2json


class Predicate(metaclass=abc.ABCMeta):
    """
    An abstract predicate representation
    """
    @abc.abstractmethod
    def __init__(self):
        pass

    @abc.abstractmethod
    def __str__(self):
        pass

    @abc.abstractmethod
    def create(tree, nodes_from, flow):
        """
        Create the predicate

        :param nodes_from: nodes which are used within edge definition
        :type nodes_from: List[Nodes]
        :param flow: flow to which predicate belongs to
        :type flow: Flow
        :return: Predicate instance
        """
        pass

    @abc.abstractmethod
    def ast(self):
        """
        :return: AST representation of predicate
        """
        pass

    @abc.abstractmethod
    def predicates_used(self):
        """
        :return: list of predicates that are used
        :rtype: List[Predicate]
        """
        pass

    @staticmethod
    def construct(tree, nodes_from, flow):
        """
        Top-down creation of predicates - recursively called to construct predicates

        :param tree: a dictionary describing nodes
        :type tree: dict
        :param nodes_from: nodes which are used within edge
        :param flow: flow to which predicate belongs to
        :type flow: Flow
        :rtype: Predicate
        """
        from .leafPredicate import LeafPredicate
        from .buildinPredicate import OrPredicate, AndPredicate, NotPredicate

        if not tree:
            raise ValueError("Bad condition '%s'" % tree)

        if 'name' in tree:
            if 'node' in tree:
                node = None
                for n in nodes_from:
                    if n.name == tree['node']:
                        node = n
                        break
                if node is None:
                    raise ValueError("Node listed node '%s' in predicate '%s' is not requested in 'nodes_from'"
                                     % (tree['node'], tree['name']))
            else:
                if len(nodes_from) > 1:
                    raise ValueError("Expected task name for predicate '%s'" % tree['name'])
                elif len(nodes_from) == 1:
                    node = nodes_from[0]
                else:
                    # e.g. starting edge has no nodes_from
                    node = None

            return LeafPredicate.create(tree['name'], node, flow, tree.get('args'))
        elif 'or' in tree:
            return OrPredicate.create(tree['or'], nodes_from, flow)
        elif 'not' in tree:
            return NotPredicate.create(tree['not'], nodes_from, flow)
        elif 'and' in tree:
            return AndPredicate.create(tree['and'], nodes_from, flow)
        else:
            raise ValueError("Unknown predicate:\n%s" % dict2json(tree))

    @abc.abstractmethod
    def check(self):
        """
        Recursively check predicate correctness

        :raises ValueError: if predicate is not correct
        """
        pass




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################


import ast
from functools import reduce
from .logger import Logger
from .predicate import Predicate


_logger = Logger.get_logger(__name__)


class BuildinPredicate(Predicate):
    """
    Build in predicate abstract class
    """
    pass


class NaryPredicate(BuildinPredicate):
    """
    N-ary predicate abstract class
    """
    def __init__(self, children):
        self._children = children

    def _str(self, op):
        ret = ""
        for child in self._children:
            if len(ret) > 0:
                ret += " %s " % op
            ret += str(child)

        if len(self._children) > 1:
            ret = "(" + ret + ")"

        return ret

    @staticmethod
    def _create(tree, cls, nodes_from, flow):
        """
        Instantiate N-ary predicate cls

        :param tree: node from which should be predicate instantiated
        :type tree: List
        :param cls: class of type NaryPredicate
        :param nodes_from: nodes that are used in described edge
        :param flow: flow to which predicate belongs to
        :type flow: Flow
        :return: instance of cls
        """
        if not isinstance(tree, list):
            raise ValueError("Nary logical operators expect list of children")
        children = []
        for child in tree:
            children.append(Predicate.construct(child, nodes_from, flow))
        return cls(children)

    def predicates_used(self):
        """
        :return: used predicates by children
        """
        return reduce(lambda x, y: x + y.predicates_used(), self._children, [])

    def check(self):
        """
        Check predicate for consistency
        """
        for child in self._children:
            child.check()


class UnaryPredicate(BuildinPredicate):
    """
    Unary predicate abstract class
    """
    def __init__(self, child):
        self._child = child

    @staticmethod
    def _create(tree, cls, nodes_from, flow):
        """
        Instantiate N-ary predicate cls

        :param tree: node from which should be predicate instantiated
        :type tree: List
        :param cls: class of type NaryPredicate
        :param nodes_from: nodes that are used in described edge
        :param flow: flow to which predicate belongs to
        :return: instance of cls
        """
        if isinstance(tree, list):
            raise ValueError("Unary logical operators expect one child")
        return cls(Predicate.construct(tree, nodes_from, flow))

    def predicates_used(self):
        """
        :return: used predicates by children
        """
        return self._child.predicates_used()

    def check(self):
        """
        Check predicate for consistency
        """
        self._child.check()


class AndPredicate(NaryPredicate):
    """
    And predicate representation
    """
    def __str__(self):
        return "(" + reduce(lambda x, y: str(x) + ' and ' + str(y), self._children) + ")"

    def ast(self):
        """
        :return: AST of describing all children predicates
        """
        return ast.BoolOp(ast.And(), [ast.Expr(value=x.ast()) for x in self._children])

    @staticmethod
    def create(tree, nodes_from, flow):
        """
        Create And predicate

        :param tree: node from which should be predicate instantiated
        :type tree: List
        :param nodes_from: nodes that are used in described edge
        :param flow: flow to which predicate belongs to
        :type flow: Flow
        :return: instance of cls
        """
        return NaryPredicate._create(tree, AndPredicate, nodes_from, flow)


class OrPredicate(NaryPredicate):
    """
    And predicate representation
    """
    def __str__(self):
        return "(" + reduce(lambda x, y: str(x) + ' or ' + str(y), self._children) + ")"

    def ast(self):
        """
        :return: AST of describing all children predicates
        """
        return ast.BoolOp(ast.Or(), [ast.Expr(value=x.ast()) for x in self._children])

    @staticmethod
    def create(tree, nodes_from, flow):
        """
        Create Or predicate

        :param tree: node from which should be predicate instantiated
        :type tree: List
        :param nodes_from: nodes that are used in described edge
        :param flow: flow to which predicate belongs to
        :type flow: Flow
        :return: instance of cls
        """
        return NaryPredicate._create(tree, OrPredicate, nodes_from, flow)


class NotPredicate(UnaryPredicate):
    """
    Unary or predicate representation
    """
    def __str__(self):
        return "(not %s)" % str(self._child)

    def ast(self):
        """
        :return: AST of describing all children predicates
        """
        return ast.UnaryOp(ast.Not(), ast.Expr(value=self._child.ast()))

    @staticmethod
    def create(tree, nodes_from, flow):
        """
        Create Or predicate

        :param tree: node from which should be predicate instantiated
        :type tree: List
        :param nodes_from: nodes that are used in described edge
        :param flow: flow to which predicate belongs to
        :type flow: Flow
        :return: instance of cls
        """
        return UnaryPredicate._create(tree, NotPredicate, nodes_from, flow)


class AlwaysTruePredicate(BuildinPredicate):
    """
    Predicate used if condition in config file is omitted
    """
    def __init__(self, flow):
        self.flow = flow

    def __str__(self):
        return "True"

    def predicates_used(self):
        return []

    def check(self):
        """
        Check predicate for consistency
        """
        pass

    def ast(self):
        """
        :return: AST
        """
        # We should return:
        #   return ast.NameConstant(value=True)
        # but it does not work with codegen
        return ast.Name(id='True', ctx=ast.Load())

    @staticmethod
    def create(tree, nodes_from, flow):
        return AlwaysTruePredicate(flow)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

from .node import Node
from .logger import Logger

_logger = Logger.get_logger(__name__)


class Flow(Node):
    """
    Flow representation
    """
    def __init__(self, name, edges=None, failures=None, nowait_nodes=None):
        """
        :param name: flow name
        :type name: str
        :param edges: edges in the flow
        :type edges: List[Edge]
        :param nowait_nodes: nodes that should not be waited for
        :type nowait_nodes: List[Node]
        """
        super(Flow, self).__init__(name)
        _logger.debug("Creating flow '{}'".format(name))
        self.edges = edges if edges else []
        self.failures = failures if failures else None
        self.nowait_nodes = nowait_nodes if nowait_nodes else []

        self.propagate_node_args = False
        self.propagate_finished = False
        self.propagate_parent = False
        self.propagate_compound_finished = False
        self.propagate_compound_parent = False

    @staticmethod
    def from_dict(d):
        raise NotImplementedError()

    def add_edge(self, edge):
        """
        :param edge: edge to be added
        :type edge: List[Edge]
        """
        self.edges.append(edge)

    def add_nowait_node(self, node):
        """
        :param node: add a node that should be marked with nowait flag
        """
        self.nowait_nodes.append(node)

    def all_nodes_from(self):
        """
        :return: all source nodes in flow, excluding failures
        """
        all_nodes_from = set()

        for edge in self.edges:
            all_nodes_from = all_nodes_from | set(edge.nodes_from)

        return list(all_nodes_from)

    def all_nodes_to(self):
        """
        :return: all destination nodes in flow, excluding failures
        """
        all_nodes_to = set()

        for edge in self.edges:
            all_nodes_to = all_nodes_to | set(edge.nodes_to)

        return list(all_nodes_to)

    def all_source_nodes(self):
        """
        :return: all source nodes in flow, including failures
        """
        if self.failures:
            return list(set(self.all_nodes_from()) | set(self.failures.all_waiting_nodes()))
        else:
            return self.all_nodes_from()

    def all_destination_nodes(self):
        """
        :return: all destination nodes in flow, including failures
        """
        if self.failures:
            return list(set(self.all_nodes_to()) | set(self.failures.all_fallback_nodes()))
        else:
            return self.all_nodes_to()

    def all_used_nodes(self):
        """
        :return: all used nodes in flow
        """
        return list(set(self.all_destination_nodes()) | set(self.all_source_nodes()))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

import codegen
import graphviz
import os
import yaml
import platform
from datetime import datetime
from .task import Task
from .storage import Storage
from .flow import Flow
from .edge import Edge
from .version import selinonlib_version
from .config import Config
from .logger import Logger
from .helpers import dict2strkwargs, expr2str
from .failures import Failures
from .taskClass import TaskClass
from .globalConfig import GlobalConfig


_logger = Logger.get_logger(__name__)


class System(object):
    """
    The representation of the whole system
    """
    def __init__(self, tasks=None, flows=None, storages=None, task_classes=None):
        self.flows = flows if flows else []
        self.tasks = tasks if tasks else []
        self.storages = storages if storages else []
        self.task_classes = task_classes if task_classes else []

    def _check_name_collision(self, name):
        """
        Tasks and Flows share name space, check for collisions

        :param name: a node name
        :raises: ValueError
        """
        if any(flow.name == name for flow in self.flows):
            raise ValueError("Unable to add node with name '%s', a flow with the same name already exist" % name)
        if any(task.name == name for task in self.tasks):
            raise ValueError("Unable to add node with name '%s', a task with the same name already exist" % name)

    def add_task(self, task):
        """
        Register a task in the system

        :param task: a task to be registered
        :type task: Task
        """
        self._check_name_collision(task.name)
        self.tasks.append(task)

    def add_flow(self, flow):
        """
        Register a flow in the system

        :param flow: a flow to be registered
        :type flow: flow
        """
        self._check_name_collision(flow.name)
        self.flows.append(flow)

    def add_storage(self, storage):
        """
        Add storage to system

        :param storage: storage that should be added
        """
        # We need to check for name collision with tasks as well since we import them by name
        for task in self.tasks:
            if task.name == storage.name:
                raise ValueError("Storage has same name as task '%s'" % storage.name)

        for stored_storage in self.storages:
            if stored_storage.name == storage:
                raise ValueError("Multiple storages of the same name '{}'".format(storage.name))
        self.storages.append(storage)

    def storage_by_name(self, name, graceful=False):
        """
        Retrieve storage by its name

        :param name: name of the storage
        :param graceful: if true, exception is raised if no such storage with name name is found
        :return: storage
        """
        for storage in self.storages:
            if storage.name == name:
                return storage
        if not graceful:
            raise KeyError("Storage with name {} not found in the system".format(name))
        return None

    def task_by_name(self, name, graceful=False):
        """
        Find a task by its name

        :param name: a task name
        :param graceful: if True, no exception is raised if task couldn't be found
        :return: task with name 'name'
        :rtype: Task
        :raises: KeyError
        """
        for task in self.tasks:
            if task.name == name:
                return task
        if not graceful:
            raise KeyError("Task with name {} not found in the system".format(name))
        return None

    def task_queue_names(self):
        """
        :return: corresponding mapping from task name to task queue
        """
        ret = {}

        for task in self.tasks:
            ret[task.name] = task.queue_name

        return ret

    def dispatcher_queue_name(self):
        """
        :return: dispatcher queue name
        """
        # leave this as method (no static/class), so we know that system is instantiated
        return GlobalConfig.dispatcher_queue

    def flow_by_name(self, name, graceful=False):
        """
        Find a flow by its name

        :param name: a flow name
        :param graceful: if True, no exception is raised if flow couldn't be found
        :return: flow with name 'name'
        :rtype: Flow
        :raises: KeyError
        """
        for flow in self.flows:
            if flow.name == name:
                return flow
        if not graceful:
            raise KeyError("Flow with name '{}' not found in the system".format(name))
        return None

    def node_by_name(self, name, graceful=False):
        """
        Find a node (flow or task) by its name

        :param name: a node name
        :param graceful: if True, no exception is raised if node couldn't be found
        :return: flow with name 'name'
        :rtype: Node
        :raises: KeyError
        """
        node = self.task_by_name(name, graceful=True)

        if not node:
            node = self.flow_by_name(name, graceful=True)

        if not node and not graceful:
            raise KeyError("Entity with name '{}' not found in the system".format(name))

        return node

    def class_of_task(self, task):
        """
        Return task class of a task

        :param task: task to look task class for
        :return: TaskClass or None if a task class for task is not available
        """
        for task_class in self.task_classes:
            if task_class.task_of_class(task):
                return task_class
        return None

    def _dump_imports(self, output):
        """
        Dump used imports of tasks to a stream

        :param output: a stream to write to
        """
        predicates = set([])
        for flow in self.flows:
            for edge in flow.edges:
                predicates.update([p.__name__ for p in edge.predicate.predicates_used()])
        if predicates:
            output.write('from %s import %s\n' % (GlobalConfig.predicates_module, ", ".join(predicates)))

        for task in self.tasks:
            output.write("from {} import {}\n".format(task.import_path, task.class_name))

        for storage in self.storages:
            if len(storage.tasks) > 0:
                output.write("from {} import {}\n".format(storage.import_path, storage.class_name))

        # we need partial for strategy function and for using storage as trace destination
        output.write("\nimport functools\n")
        output.write("from {} import {} as _strategy_function\n".format(GlobalConfig.strategy_module,
                                                                        GlobalConfig.strategy_function))
        output.write('\n\n')

    def _dump_is_flow(self, output):
        """
        Dump is_flow() check to a stream

        :param output: a stream to write to
        """
        output.write('def is_flow(name):\n')
        output.write('    return name in %s\n\n' % str([flow.name for flow in self.flows]))

    def _dump_output_schemas(self, output):
        """
        Dump output schema mapping to a stream

        :param output: a stream to write to
        """
        output.write('output_schemas = {')
        printed = False
        for task in self.tasks:
            if task.output_schema:
                if printed:
                    output.write(",")
                output.write("\n    '%s': '%s'" % (task.name, task.output_schema))
                printed = True
        output.write('\n}\n\n')

    def _dump_dict(self, output, dict_name, dict_items):
        """
        Dump propagate_finished flag configuration to a stream

        :param output: a stream to write to
        """
        output.write('%s = {' % dict_name)
        printed = False
        for item in dict_items:
            for key, value in item.items():
                if printed:
                    output.write(",")
                if isinstance(value, list):
                    string = [item.name for item in value]
                else:
                    string = str(value)
                output.write("\n    '%s': %s" % (key, string))
                printed = True
        output.write('\n}\n\n')

    def _dump_task_classes(self, output):
        """
        Dump mapping from task name to task class

        :param output: a stream to write to
        """
        output.write('task_classes = {')
        printed = False
        for task in self.tasks:
            if printed:
                output.write(',')
            output.write("\n    '%s': %s" % (task.name, task.class_name))
            printed = True
        output.write('\n}\n\n')

    def _dump_queues(self, output):
        output.write('task_queues = {')
        printed = False
        for task in self.tasks:
            if printed:
                output.write(',')
            output.write("\n    '%s': '%s'" % (task.name, task.queue_name))
            printed = True
        output.write('\n}\n\n')

        output.write("dispatcher_queue = '%s'\n\n" % self.dispatcher_queue_name())

    @staticmethod
    def _dump_get_task_instance(output):
        """
        Dump get_task_instance() function to a stream

        :param output: a stream to write to
        """
        output.write('def get_task_instance(task_name, flow_name, parent, finished):\n')
        output.write("    cls = task_classes.get(task_name)\n")
        output.write("    if not cls:\n")
        output.write("        raise ValueError(\"Unknown task with name '%s'\" % flow_name)\n")
        output.write("    return cls(task_name=task_name, flow_name=flow_name, parent=parent, finished=finished)\n\n")

    def _dump_storage2instance_mapping(self, output):
        """
        Dump storage name to instance mapping to a stream

        :param output: a stream to write to
        """
        storage_var_names = []
        for storage in self.storages:
            if len(storage.tasks) > 0:
                output.write("%s = %s" % (storage.var_name, storage.class_name))
                if storage.configuration and isinstance(storage.configuration, dict):
                    output.write("(%s)\n" % dict2strkwargs(storage.configuration))
                elif storage.configuration:
                    output.write("(%s)\n" % expr2str(storage.configuration))
                else:
                    output.write("()\n")

                storage_var_names.append((storage.name, storage.var_name,))

        output.write('storage2instance_mapping = {\n')
        printed = False
        for storage_var_name in storage_var_names:
                if printed:
                    output.write(",\n")
                output.write("    '%s': %s" % (storage_var_name[0], storage_var_name[1]))
                printed = True

        output.write("\n}\n\n")

    def _dump_task2storage_mapping(self, output):
        """
        Dump task name to storage name mapping to a stream

        :param output: a stream to write to
        """
        output.write('task2storage_mapping = {\n')
        printed = False
        for storage in self.storages:
            for task in storage.tasks:
                if printed:
                    output.write(",\n")
                output.write("    '%s': '%s'" % (task.name, storage.name))
                printed = True
        output.write("\n}\n\n")

    @staticmethod
    def _dump_strategy_func(output):
        """
        Dump scheduling strategy function to a stream

        :param output: a stream to write to
        """
        # functools is imported in self._dump_imports() so it is safe to use
        output.write('strategy_function = functools.partial(_strategy_function, %s)\n\n'
                     % dict2strkwargs(GlobalConfig.strategy_func_args))

    @staticmethod
    def _dump_condition_name(flow_name, idx):
        """
        Create condition name for a dump

        :param flow_name: flow name
        :type flow_name: str
        :param idx: index of condition within the flow
        :type idx: int
        :return: condition function representation
        """
        assert(idx >= 0)
        return '_condition_{}_{}'.format(flow_name, idx)

    def _dump_condition_functions(self, output):
        """
        Dump condition functions to a stream

        :param output: a stream to write to
        """
        for flow in self.flows:
            for idx, edge in enumerate(flow.edges):
                output.write('def {}(db, node_args):\n'.format(self._dump_condition_name(flow.name, idx)))
                output.write('    return {}\n\n\n'.format(codegen.to_source(edge.predicate.ast())))

    def _dump_max_retry(self, output):
        """
        Dump max_retry configuration to a stream

        :param output: a stream to write to
        """
        output.write('max_retry = {')
        printed = False
        for task in self.tasks:
            if printed:
                output.write(',')
            output.write("\n    '%s': %d" % (task.name, task.max_retry))
            printed = True
        output.write('\n}\n\n')

    def _dump_retry_countdown(self, output):
        """
        Dump retry_countdown configuration to a stream

        :param output: a stream to write to
        """
        output.write('retry_countdown = {')
        printed = False
        for task in self.tasks:
            if printed:
                output.write(',')
            output.write("\n    '%s': %d" % (task.name, task.retry_countdown))
            printed = True
        output.write('\n}\n\n')

    def _dump_storage_readonly(self, output):
        """
        Dump storage_readonly flow to a stream

        :param output: a stream to write to
        """
        output.write('storage_readonly = {')
        printed = False
        for task in self.tasks:
            if printed:
                output.write(',')
            output.write("\n    '%s': %s" % (task.name, task.storage_readonly))
            printed = True
        output.write('\n}\n\n')

    def _dump_time_limit(self, output):
        """
        Dump max_retry configuration to a stream

        :param output: a stream to write to
        """
        output.write('time_limit = {')
        printed = False
        for task in self.tasks:
            if printed:
                output.write(',')
            output.write("\n    '%s': %s" % (task.name, task.time_limit))
            printed = True
        output.write('\n}\n\n')

    def _dump_nowait_nodes(self, output):
        """
        Dump nowait nodes to a stream

        :param output: a stream to write to
        """

        output.write('nowait_nodes = {\n')
        printed = False
        for flow in self.flows:
            if printed:
                    output.write(',\n')
            output.write("    '%s': %s" % (flow.name, [node.name for node in flow.nowait_nodes]))
            printed = True

        output.write('\n}\n\n')

    @staticmethod
    def _dump_init(output):
        """
        Dump init function to a stream

        :param output: a stream to write to
        :return:
        """
        output.write('def init(config_cls):\n')
        GlobalConfig.dump_trace(output, 'config_cls', indent_count=1)
        # always pass in case we have nothing to init
        output.write('    return\n')
        output.write('\n')

    def _dump_edge_table(self, output):
        """
        Dump edge definition table to a stream

        :param output: a stream to write to
        """
        output.write('edge_table = {\n')
        for idx, flow in enumerate(self.flows):
            output.write("    '{}': [".format(flow.name))
            for idx_edge, edge in enumerate(flow.edges):
                if idx_edge > 0:
                    output.write(',\n')
                    output.write(' '*(len(flow.name) + 4 + 5)) # align to previous line
                output.write("{'from': %s" % str([node.name for node in edge.nodes_from]))
                output.write(", 'to': %s" % str([node.name for node in edge.nodes_to]))
                output.write(", 'condition': %s}" % self._dump_condition_name(flow.name, idx_edge))
            if idx + 1 < len(self.flows):
                output.write('],\n')
            else:
                output.write(']\n')
        output.write('}\n\n')

    def dump2stream(self, f):
        """
        Perform system dump to a Python source code to an output stream

        :param f: an output stream to write to
        """
        f.write('#!/usr/bin/env python3\n')
        f.write('# auto-generated using Selinonlib v{} on {} at {}\n\n'.format(selinonlib_version,
                                                                            platform.node(),
                                                                            str(datetime.utcnow())))
        self._dump_imports(f)
        self._dump_task_classes(f)
        self._dump_queues(f)
        self._dump_get_task_instance(f)
        f.write('#'*80+'\n\n')
        self._dump_task2storage_mapping(f)
        self._dump_storage2instance_mapping(f)
        f.write('#'*80+'\n\n')
        self._dump_is_flow(f)
        f.write('#'*80+'\n\n')
        self._dump_strategy_func(f)
        f.write('#'*80+'\n\n')
        self._dump_output_schemas(f)
        f.write('#'*80+'\n\n')
        self._dump_dict(f, 'propagate_node_args', [{f.name: f.propagate_node_args} for f in self.flows])
        self._dump_dict(f, 'propagate_finished', [{f.name: f.propagate_finished} for f in self.flows])
        self._dump_dict(f, 'propagate_parent', [{f.name: f.propagate_parent} for f in self.flows])
        self._dump_dict(f, 'propagate_compound_finished', [{f.name: f.propagate_compound_finished} for f in self.flows])
        self._dump_dict(f, 'propagate_compound_parent', [{f.name: f.propagate_compound_parent} for f in self.flows])
        f.write('#'*80+'\n\n')
        self._dump_max_retry(f)
        self._dump_retry_countdown(f)
        self._dump_time_limit(f)
        self._dump_storage_readonly(f)
        f.write('#'*80+'\n\n')

        for flow in self.flows:
            if flow.failures:
                flow.failures.dump2stream(f, flow.name)

        f.write('failures = {')
        printed = False
        for i, flow in enumerate(self.flows):
            if flow.failures:
                if printed:
                    f.write(",")
                printed = True
                f.write("\n    '%s': %s" % (flow.name, flow.failures.starting_nodes_name(flow.name)))
        f.write('\n}\n\n')

        f.write('#'*80+'\n\n')
        self._dump_nowait_nodes(f)
        f.write('#'*80+'\n\n')
        self._dump_init(f)
        f.write('#'*80+'\n\n')
        self._dump_condition_functions(f)
        f.write('#'*80+'\n\n')
        self._dump_edge_table(f)
        f.write('#'*80+'\n\n')

    def dump2file(self, output_file):
        """
        Perform system dump to a Python source code

        :param output_file: an output file to write to
        """
        _logger.debug("Performing system dump to '%s'" % output_file)
        with open(output_file, 'w') as f:
            self.dump2stream(f)

    def plot_graph(self, output_dir, image_format=None):
        """
        Plot system flows to graphs - each flow in a separate file

        :param output_dir: output directory to write graphs of flows to
        :param image_format: image format, the default is svg if None
        :return: list of file names to which the graph was rendered
        :rtype: List[str]
        """
        _logger.debug("Rendering system flows to '%s'" % output_dir)
        ret = []
        image_format = image_format if image_format else 'svg'

        for flow in self.flows:
            storage_connections = []
            graph = graphviz.Digraph(format=image_format)
            graph.graph_attr.update(Config().style_graph())
            graph.node_attr.update(Config().style_task())
            graph.edge_attr.update(Config().style_edge())

            for idx, edge in enumerate(flow.edges):
                condition_node = "%s_%s" % (flow.name, idx)
                graph.node(name=condition_node, label=str(edge.predicate), _attributes=Config().style_condition())

                for node in edge.nodes_to:
                    # Plot storage connection
                    if node.is_flow():
                        graph.node(name=node.name, _attributes=Config().style_flow())
                    else:
                        graph.node(name=node.name)
                        if node.storage:
                            graph.node(name=node.storage.name, _attributes=Config().style_storage())
                            if (node.name, node.storage.name) not in storage_connections:
                                graph.edge(node.name, node.storage.name, _attributes=Config().style_store_edge())
                                storage_connections.append((node.name, node.storage.name,))
                    graph.edge(condition_node, node.name)
                for node in edge.nodes_from:
                    if node.is_flow():
                        graph.node(name=node.name, _attributes=Config().style_flow())
                    else:
                        graph.node(name=node.name)
                        if node.storage:
                            graph.node(name=node.storage.name, _attributes=Config().style_storage())
                            if (node.name, node.storage.name) not in storage_connections:
                                graph.edge(node.name, node.storage.name, _attributes=Config().style_store_edge())
                                storage_connections.append((node.name, node.storage.name,))
                    graph.edge(node.name, condition_node)

            # Plot failures as well
            if flow.failures:
                for failure in flow.failures.raw_definition:
                    if len(failure['nodes']) == 1 and \
                            (isinstance(failure['fallback'], bool) or len(failure['fallback']) == 1):
                        graph.node(name=failure['nodes'][0])

                        if isinstance(failure['fallback'], list):
                            fallback_node_name = failure['fallback'][0]
                        else:
                            fallback_node_name = str(failure['fallback'])

                        graph.node(name=fallback_node_name)
                        graph.edge(failure['nodes'][0], fallback_node_name,
                                   _attributes=Config().style_fallback_edge())
                    else:
                        graph.node(name=str(id(failure)), _attributes=Config().style_fallback_node())

                        for node_name in failure['nodes']:
                            if self.node_by_name(node_name).is_flow():
                                graph.node(name=node_name, _attributes=Config().style_flow())
                            else:
                                graph.node(name=node_name)
                            graph.edge(node_name, str(id(failure)), _attributes=Config().style_fallback_edge())

                        if failure['fallback'] is True:
                            graph.node(name=str(id(failure['fallback'])), label="True",
                                       _attributes=Config().style_fallback_true())
                            graph.edge(str(id(failure)), str(id(failure['fallback'])),
                                       _attributes=Config().style_fallback_edge())
                        else:
                            for node_name in failure['fallback']:
                                if self.node_by_name(node_name).is_flow():
                                    graph.node(name=node_name, _attributes=Config().style_flow())
                                else:
                                    graph.node(name=node_name)
                                graph.edge(str(id(failure)), node_name, _attributes=Config().style_fallback_edge())

            file = os.path.join(output_dir, "%s" % flow.name)
            graph.render(filename=file, cleanup=True)
            ret.append(file)
            _logger.info("Graph rendered to '%s.%s'" % (file, image_format))

        return ret

    def _post_parse_check(self):
        """
        Called once parse was done to ensure that system was correctly defined in config file

        :raises: ValueError
        """
        _logger.debug("Post parse check is going to be executed")
        # we want to have circular dependencies, so we need to check consistency after parsing since all flows
        # are listed (by names) in a separate definition
        for flow in self.flows:
            if len(flow.edges) == 0:
                raise ValueError("Empty flow: %s" % flow.name)

    def _check(self):
        """
        Check system for consistency

        :raises: ValueError
        """
        _logger.info("Checking system for consistency")

        for task_class in self.task_classes:
            task_ref = task_class.tasks[0]
            for task in task_class.tasks[1:]:
                if task_ref.output_schema != task.output_schema:
                    _logger.warning("Different output schemas to a same task class: %s and %s for class '%s', "
                                    "schemas: '%s' and '%s' might differ"
                                    % (task_ref.name, task.name, task_class.class_name,
                                       task_ref.output_schema, task.output_schema))

                if task.max_retry != task_ref.max_retry:
                    _logger.warning("Different max_retry to a same task class: %s and %s for class '%s'"
                                    % ((task.name, task.max_retry), (task_ref.name, task_ref.max_retry),
                                       task_class.class_name))

                if task.time_limit != task_ref.time_limit:
                    _logger.warning("Different time_limit to a same task class: %s and %s for class '%s'"
                                    % ((task.name, task.max_retry), (task_ref.name, task_ref.max_retry),
                                       task_class.class_name))

        for storage in self.storages:
            if len(storage.tasks) == 0:
                _logger.warning("Storage '{}' not used in any flow".format(storage.name))

        # We want to check that if we depend on a node, that node is being started at least once in the flow
        # This also covers check for starting node definition
        for flow in self.flows:
            # TODO: we should make this more transparent by placing it to separate functions
            try:
                all_source_nodes = flow.all_source_nodes()
                all_destination_nodes = flow.all_destination_nodes()

                starting_nodes_count = 0

                for edge in flow.edges:
                    if len(edge.nodes_from) == 0:
                        starting_nodes_count += 1

                    node_seen = {}
                    for node_from in edge.nodes_from:
                        if not node_seen.get(node_from.name, False):
                            node_seen[node_from.name] = True
                        else:
                            raise ValueError("Nodes cannot be dependent on a node of a same type mode than once; "
                                             "node from '%s' more than once in flow '%s'" % (node_from.name, flow.name))

                    edge.predicate.check()

                if starting_nodes_count > 1:
                    _logger.warning("Multiple starting nodes defined in flow '%s'" % flow.name)

                if starting_nodes_count == 0:
                    raise ValueError("No starting node found in flow '%s'" % flow.name)

                for nowait_node in flow.nowait_nodes:
                    if nowait_node in all_source_nodes:
                        raise ValueError("Node '%s' marked as 'nowait' but dependency in the flow found"
                                         % nowait_node.name)

                if isinstance(flow.propagate_finished, list):
                    for node in flow.propagate_finished:
                        if node not in all_source_nodes:
                            raise ValueError("Subflow '%s' should receive parent nodes, but there is no dependency "
                                             "in flow '%s' to which should be parent nodes propagated"
                                             % (node.name, flow.name))

                if isinstance(flow.propagate_compound_finished, list):
                    for node in flow.propagate_compound_finished:
                        if node not in all_source_nodes:
                            raise ValueError("Subflow '%s' should receive parent nodes, but there is no dependency "
                                             "in flow '%s' to which should be parent nodes propagated"
                                             % (node.name, flow.name))

                if isinstance(flow.propagate_parent, list):
                    for node in flow.propagate_parent:
                        if node not in all_source_nodes:
                            raise ValueError("Subflow '%s' should receive parent, but there is no dependency "
                                             "in flow '%s' to which should be parent nodes propagated"
                                             % (node.name, flow.name))

                if isinstance(flow.propagate_compound_parent, list):
                    for node in flow.propagate_compound_parent:
                        if node not in all_source_nodes:
                            raise ValueError("Subflow '%s' should receive parent, but there is no dependency "
                                             "in flow '%s' to which should be parent nodes propagated"
                                             % (node.name, flow.name))

                if isinstance(flow.propagate_finished, list) and isinstance(flow.propagate_compound_finished, list):
                    for node in flow.propagate_finished:
                        if node in flow.propagate_compound_finished:
                            raise ValueError("Cannot mark node '%s' for propagate_finished and "
                                             "propagate_compound_finished at the same time in flow '%s'"
                                             % (node.name, flow.name))

                if isinstance(flow.propagate_parent, list) and isinstance(flow.propagate_compound_parent, list):
                    for node in flow.propagate_parent:
                        if node in flow.propagate_compound_parent:
                            raise ValueError("Cannot mark node '%s' for propagate_parent and propagate_compound_parent "
                                             "at the same time in flow '%s'" % (node.name, flow.name))

                if isinstance(flow.propagate_parent, list) and flow.propagate_compound_parent is True or \
                        flow.propagate_parent is True and isinstance(flow.propagate_compound_parent, list):
                    raise ValueError("Misleading configuration of propagate_parent and "
                                     "propagate_compound_parent for flow '%s'" % flow.name)

                if isinstance(flow.propagate_finished, list) and flow.propagate_compound_finished is True or \
                        flow.propagate_finished is True and isinstance(flow.propagate_compound_finished, list):
                    raise ValueError("Misleading configuration of propagate_finished and "
                                     "propagate_compound_finished for flow '%s'" % flow.name)

                not_started = list(set(all_source_nodes) - set(all_destination_nodes))

                error = False
                for node in not_started:
                    if node.is_task():
                        _logger.error("Dependency in flow '%s' on node '%s', but this node is not started in the flow"
                                      % (flow.name, node.name))
                        error = True

                if error:
                    raise ValueError("Dependency on not started node detected in flow '%s'" % flow.name)

                if self.dispatcher_queue_name() in self.task_queue_names().keys():
                    raise ValueError("Detected collision in queue names - dispatcher queue name %s, "
                                     "task queue names: %s" % (self.dispatcher_queue_name(), self.task_queue_names()))
            except:
                _logger.error("Check of flow '%s' failed" % flow.name)
                raise

    @staticmethod
    def _set_propagate(system, flow_name, flow_def, propagate_type):
        """
        Parse propagate_node_args flag and adjust flow accordingly

        :param system: system that is used
        :param flow_name: flow name for which flag is configured
        :param flow_def: flow definition
        :param propagate_type: propagate flag type
        """
        ret = False

        if propagate_type in flow_def and flow_def[propagate_type] is not None:
            if not isinstance(flow_def[propagate_type], list) and \
                    not isinstance(flow_def[propagate_type], bool):
                flow_def[propagate_type] = [flow_def[propagate_type]]

            if isinstance(flow_def[propagate_type], list):
                ret = []
                for node_name in flow_def[propagate_type]:
                    node = system.flow_by_name(node_name)
                    ret.append(node)
            elif isinstance(flow_def[propagate_type], bool):
                ret = flow_def[propagate_type]
            else:
                raise ValueError("Unknown value in '%s' in flow %s" % (flow_name, propagate_type))

        return ret

    @staticmethod
    def from_files(nodes_definition_file, flow_definition_files, no_check=False):
        """
        Construct System from files

        :param nodes_definition_file: path to nodes definition file
        :param flow_definition_files: path to files that describe flows
        :param no_check: True if system shouldn't be checked for consistency (recommended to check)
        :return: System instance
        :rtype: System
        """
        system = System()

        with open(nodes_definition_file, 'r') as f:
            _logger.debug("Parsing '{}'".format(nodes_definition_file))
            try:
                content = yaml.load(f, Loader=yaml.SafeLoader)
            except:
                _logger.error("Bad YAML file, unable to load tasks from {}".format(nodes_definition_file))
                raise

        for storage_dict in content.get('storages', []):
            storage = Storage.from_dict(storage_dict)
            system.add_storage(storage)

        if 'global' in content:
            GlobalConfig.from_dict(system, content['global'])

        if 'tasks' not in content or content['tasks'] is None:
            raise ValueError("No tasks defined in the system")

        for task_dict in content['tasks']:
            task = Task.from_dict(task_dict, system)
            task_class = system.class_of_task(task)
            if not task_class:
                task_class = TaskClass(task.class_name, task.import_path)
                system.task_classes.append(task_class)
            task_class.add_task(task)
            task.task_class = task_class
            system.add_task(task)

        for flow_name in content['flows']:
            flow = Flow(flow_name)
            system.add_flow(flow)

        if not isinstance(flow_definition_files, list):
            flow_definition_files = [flow_definition_files]

        for flow_file in flow_definition_files:
            with open(flow_file, 'r') as f:
                _logger.debug("Parsing '{}'".format(flow_file))
                try:
                    content = yaml.load(f, Loader=yaml.SafeLoader)
                except:
                    _logger.error("Bad YAML file, unable to load flow from {}".format(flow_file))
                    raise

            for flow_def in content['flow-definitions']:
                flow = system.flow_by_name(flow_def['name'])

                if len(flow.edges) > 0:
                    raise ValueError("Multiple definitions of flow '%s'" % flow.name)

                for edge_def in flow_def['edges']:
                    edge = Edge.from_dict(edge_def, system, flow)
                    flow.add_edge(edge)

                if 'failures' in flow_def:
                    failures = Failures.construct(system, flow, flow_def['failures'])
                    flow.failures = failures

                if 'nowait' in flow_def and flow_def['nowait'] is not None:
                    if not isinstance(flow_def['nowait'], list):
                        flow_def['nowait'] = [flow_def['nowait']]

                    for node_name in flow_def['nowait']:
                        node = system.node_by_name(node_name)
                        flow.add_nowait_node(node)

                flow.propagate_node_args = System._set_propagate(system, flow.name, flow_def, 'propagate_node_args')
                flow.propagate_finished = System._set_propagate(system, flow.name, flow_def, 'propagate_finished')
                flow.propagate_parent = System._set_propagate(system, flow.name, flow_def, 'propagate_parent')
                flow.propagate_compound_finished = System._set_propagate(system, flow.name, flow_def,
                                                                         'propagate_compound_finished')
                flow.propagate_compound_parent = System._set_propagate(system, flow.name, flow_def,
                                                                       'propagate_compound_parent')

        system._post_parse_check()
        if not no_check:
            system._check()

        return system



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

import abc
import re


class Node(metaclass=abc.ABCMeta):
    """
    An abstract class for node representation
    """
    def __init__(self, name):
        if not self.check_name(name):
            raise ValueError("Invalid node name '%s'" % name)
        self._name = name

    @abc.abstractstaticmethod
    def from_dict(d):
        """
        Construct node from a dict

        :return: instantiated node
        """
        pass

    @property
    def name(self):
        """
        :return: a name of the node
        """
        return self._name

    def is_flow(self):
        """
        :return: True if node represents a Flow
        """
        from .flow import Flow
        return isinstance(self, Flow)

    def is_task(self):
        """
        :return: True if node represents a Task
        """
        from .task import Task
        return isinstance(self, Task)

    @staticmethod
    def check_name(name):
        """
        Check whether name is a correct node (flow/task) name

        :param name: node name
        :return: True if name is a correct node name
        :rtype: bool
        """
        r = re.compile(r"^[_a-zA-Z][_a-zA-Z0-9]*$")
        return r.match(name)



#!/usr/bin/env python3
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################


class TaskClass(object):
    """
    Task's class abstraction
    """
    def __init__(self, class_name, import_path):
        self.class_name = class_name
        self.import_path = import_path
        self.tasks = []

    def task_of_class(self, task):
        """
        :param task: task to be checked
        :return: True if task is of this class
        """
        return self.class_name == task.class_name and self.import_path == task.import_path

    def add_task(self, task):
        """
        Register a task to this class

        :param task: task to be added
        """
        assert(self.task_of_class(task))
        self.tasks.append(task)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

import os
import sys
import ast
import json
from dill.source import getsource
from contextlib import contextmanager


def dict2strkwargs(d):
    """
    Convert a dictionary into arguments to a string representation that can be used as arguments to a function
    """
    ret = ""
    for k, v in d.items():
        if len(ret) > 0:
            ret += ", "
        ret += "%s=%s" % (k, expr2str(v))
    return ret


def expr2str(e):
    """
    Convert a Python expression into a Python code
    """
    if isinstance(e, dict):
        return str(e)
    elif isinstance(e, list):
        # s/'['foo']['bar']'/['foo']['bar']/ (get rid of leading ')
        return "%s" % e
    elif isinstance(e, str):
        return "'%s'" % e
    else:
        # some build in type such as bool/int/...
        return "%s" % str(e)


def keylist2str(keylist):
    """
    Convert keylist to a string representation

    :param keylist: keylist to be converted
    :type keylist: list
    :return: string representation
    :rtype: str
    """
    return "".join(map(lambda x: "['" + str(x) + "']", keylist))


@contextmanager
def pushd(new_dir):
    """
    Traverse directory tree in push/pop manner

    :param new_dir: new directory to cd to
    :type new_dir: str
    """
    prev_dir = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(prev_dir)


def dict2json(o, pretty=True):
    """
    Convert dict to json (string)

    :param o: dictionary to be converted
    :param pretty: if True, nice formatting will be used
    :type pretty: bool
    :return: formatted dict in json
    :rtype: str
    """
    if pretty is True:
        return json.dumps(o, sort_keys=True, separators=(',', ': '), indent = 2)
    else:
        return json.dumps(o)


def get_function_arguments(function):
    """
    Get arguments of function

    :param function: function to parse arguments
    :return: list of arguments that predicate function expects
    """
    ret = []

    func_source = getsource(function)
    func_ast = ast.parse(func_source)
    call = func_ast.body[0]

    for arg in call.args.args:
        # Python2 and Python3 have different AST representations for arg, this can be removed since we support
        # only python3
        if sys.version_info[0] == 2:
            ret.append(arg.id)
        else:
            ret.append(arg.arg)

    return ret




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

from .node import Node
from .logger import Logger
from .globalConfig import GlobalConfig

_logger = Logger.get_logger(__name__)

_DEFAULT_MAX_RETRY = 0
_DEFAULT_RETRY_COUNTDOWN = 0


class Task(Node):
    """
    A task representation within the system
    """
    def __init__(self, name, import_path, storage, opts):
        """
        :param name: name of the task
        :param import_path: tasks's import
        :param storage: storage that should be used
        :param opts: additional options for task
        """
        super(Task, self).__init__(name)

        self.class_name = opts.get('classname', name)
        self.output_schema = opts.get('output_schema')
        self.max_retry = opts.get('max_retry', _DEFAULT_MAX_RETRY)
        self.time_limit = opts.get('time_limit')
        self.retry_countdown = opts.get('retry_countdown', _DEFAULT_RETRY_COUNTDOWN)
        self.queue_name = opts.get('queue', GlobalConfig.default_task_queue)
        self.import_path = import_path
        self.storage = storage
        self.storage_readonly = opts.get('storage_readonly', False)

        # register task usage
        if self.storage:
            storage.register_task(self)
        _logger.debug("Creating task with name '%s' import path '%s', class name '%s'"
                      % (self.name, self.import_path, self.class_name))

    def check(self):
        """
        Check task definitions for errors

        :raises: ValueError if an error occurred
        """
        if not isinstance(self.import_path, str):
            raise ValueError("Error in task '%s' definition - import path should be string; got '%s'"
                             % (self.name, self.import_path))

        if self.class_name is not None and not isinstance(self.class_name, str):
            raise ValueError("Error in task '%s' definition - class instance should be string; got '%s'"
                             % (self.name, self.class_name))

        if self.output_schema is not None and not isinstance(self.output_schema, str):
            raise ValueError("Error in task '%s' definition - output schema should be string; got '%s'"
                             % (self.name, self.output_schema))

        if self.max_retry is not None and (not isinstance(self.max_retry, int) or self.max_retry < 0):
            raise ValueError("Error in task '%s' definition - max_retry should be None, zero or positive integer;"
                             " got '%s'" % (self.name, self.max_retry))

        if self.time_limit is not None and (not isinstance(self.time_limit, int) or self.time_limit <= 0):
            raise ValueError("Error in task '%s' definition - time_limit should be None or positive integer;"
                             " got '%s'" % (self.name, self.time_limit))

        if self.retry_countdown is not None and (not isinstance(self.retry_countdown, int) or self.retry_countdown < 0):
            raise ValueError("Error in task '%s' definition - retry_countdown should be None or positive integer;"
                             " got '%s'" % (self.name, self.retry_countdown))

        if self.queue_name is not None and not isinstance(self.queue_name, str):
            raise ValueError("Invalid task queue, should be string, got %s" % self.queue_name)

        if not isinstance(self.storage_readonly, bool):
            raise ValueError("Storage usage flag readonly should be of type bool")

    @staticmethod
    def from_dict(d, system):
        """
        Construct task from a dict and check task's definition correctness

        :param d: dictionary to be used to construct the task
        :return: Task instance
        :param system: system that should be used to for lookup a storage
        :type system: System
        :rtype: Task
        :raises: ValueError
        """
        if 'name' not in d or not d['name']:
            raise KeyError('Task name definition is mandatory')
        if 'import' not in d or not d['import']:
            raise KeyError('Task import definition is mandatory')
        if 'storage' in d:
            storage = system.storage_by_name(d['storage'])
        else:
            storage = None

        instance = Task(d['name'], d['import'], storage, d)
        instance.check()
        return instance



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################
"""
Functions for scheduling dispatcher - Dispatcher strategy configuration functions. All of them are receiving:

  * previous_retry - value of previous retry, None if scheduling for the first time
  * active_nodes - active nodes within the system
  * failed_nodes - failed nodes within the system
  * new_started_nodes - newly started nodes the system, these nodes are already in active_nodes
  * new_fallback_nodes - newly started fallback nodes, these nodes are already in active_nodes

Other parameters are specific for strategy function and are configurable from YAML configuration file.
"""

from random import randint as gen_random


def linear_increase(start_retry, max_retry, step,
                    previous_retry, active_nodes, failed_nodes, new_started_nodes, new_fallback_nodes):
    """
    Increase linearly if no node started, decrease drastically to start_retry if a node scheduled

    :param start_retry: starting retry to use
    :param max_retry: upper limit of scheduling
    :param step: step to use in linear increase
    """
    if len(active_nodes) == 0:
        return None

    if len(new_started_nodes) > 0 or len(new_fallback_nodes) > 0:
        retry = previous_retry + step
        if retry > max_retry:
            return max_retry
        else:
            return retry
    else:
        return start_retry


def linear_adopt(start_retry, max_retry, step,
                 previous_retry, active_nodes, failed_nodes, new_started_nodes, new_fallback_nodes):
    """
    Increase linearly if no node started, decrease linearly if a node scheduled

    :param start_retry: starting retry to use
    :param max_retry: upper limit of scheduling
    :param step: step to use in linear increase
    """
    if len(active_nodes) == 0:
        return None

    if len(new_started_nodes) > 0 or len(new_fallback_nodes) > 0:
        retry = previous_retry + step
        if retry > max_retry:
            return max_retry
        else:
            return retry
    else:
        retry = previous_retry - step
        if retry < start_retry:
            return start_retry
        else:
            return retry


def biexponential_increase(start_retry, max_retry,
                           previous_retry, active_nodes, failed_nodes, new_started_nodes, new_fallback_nodes):
    """
    Increase exponentially if no node started, decrease drastically to start_retry if a node scheduled

    :param start_retry: starting retry to use
    :param max_retry: upper limit of scheduling
    """
    if previous_retry is None:
        return start_retry

    if len(active_nodes) == 0:
        return None

    if len(new_started_nodes) > 0 or len(new_fallback_nodes) > 0:
        retry = previous_retry * 2
        if retry > max_retry:
            return max_retry
        else:
            return retry
    else:
        return start_retry


def biexponential_adopt(start_retry, max_retry,
                        previous_retry, active_nodes, failed_nodes, new_started_nodes, new_fallback_nodes):
    """
    Increase exponentially if no node started, decrease exponentially if a node scheduled

    :param start_retry: starting retry to use
    :param max_retry: upper limit of scheduling
    """
    if previous_retry is None:
        return start_retry

    if len(active_nodes) == 0:
        return None

    if len(new_started_nodes) > 0 or len(new_fallback_nodes) > 0:
        retry = previous_retry * 2
        if retry > max_retry:
            return max_retry
        else:
            return retry
    else:
        retry = previous_retry / 2
        if retry < start_retry:
            return start_retry
        else:
            return retry


def random(start_retry, max_retry, previous_retry, active_nodes, failed_nodes, new_started_nodes, new_fallback_nodes):
    """
    Schedule randomly
    :param start_retry: lower limit of scheduling
    :param max_retry: upper limit of scheduling
    """
    if len(active_nodes) == 0:
        return None

    return gen_random(start_retry, max_retry)






#!/usr/bin/env python3

from http.client import HTTPConnection


def httpStatus(host, path, status):
    try:
        conn = HTTPConnection(host)
        conn.request("HEAD", path)
        return conn.getresponse().status == status
    except:
        return False



#!/usr/bin/env python3

from .alwaysFalse import alwaysFalse
from .alwaysTrue import alwaysTrue
from .argsEmpty import argsEmpty
from .argsFieldBool import argsFieldBool
from .argsFieldContain import argsFieldContain
from .argsFieldDict import argsFieldDict
from .argsFieldEqual import argsFieldEqual
from .argsFieldExist import argsFieldExist
from .argsFieldFloat import argsFieldFloat
from .argsFieldGreaterEqual import argsFieldGreaterEqual
from .argsFieldGreater import argsFieldGreater
from .argsFieldInt import argsFieldInt
from .argsFieldLenEqual import argsFieldLenEqual
from .argsFieldLenGreaterEqual import argsFieldLenGreaterEqual
from .argsFieldLenGreater import argsFieldLenGreater
from .argsFieldLenLessEqual import argsFieldLenGreaterEqual
from .argsFieldLenLess import argsFieldLenLess
from .argsFieldLenNotEqual import argsFieldLenNotEqual
from .argsFieldLessEqual import argsFieldLessEqual
from .argsFieldLess import argsFieldLess
from .argsFieldList import argsFieldList
from .argsFieldNone import argsFieldNone
from .argsFieldNotEqual import argsFieldNotEqual
from .argsFieldStr import argsFieldStr
from .argsFieldUrlNetloc import argsFieldUrlNetloc
from .argsFieldUrlPath import argsFieldUrlPath
from .argsFieldUrlScheme import argsFieldUrlScheme
from .argsIsBool import argsIsBool
from .argsIsDict import argsIsDict
from .argsIsFloat import argsIsFloat
from .argsIsInt import argsIsInt
from .argsIsList import argsIsList
from .argsIsNone import argsIsNone
from .argsIsStr import argsIsStr
from .empty import empty
from .fieldBool import fieldBool
from .fieldContain import fieldContain
from .fieldDict import fieldDict
from .fieldEqual import fieldEqual
from .fieldExist import fieldExist
from .fieldFloat import fieldFloat
from .fieldGreaterEqual import fieldGreaterEqual
from .fieldGreater import fieldGreater
from .fieldInt import fieldInt
from .fieldLenEqual import fieldLenEqual
from .fieldLenGreaterEqual import fieldLenGreaterEqual
from .fieldLenGreater import fieldLenGreater
from .fieldLenLessEqual import fieldLenGreaterEqual
from .fieldLenLess import fieldLenLess
from .fieldLenNotEqual import fieldLenNotEqual
from .fieldLessEqual import fieldLessEqual
from .fieldLess import fieldLess
from .fieldList import fieldList
from .fieldNone import fieldNone
from .fieldNotEqual import fieldNotEqual
from .fieldStr import fieldStr
from .fieldUrlNetloc import fieldUrlNetloc
from .fieldUrlPath import fieldUrlPath
from .fieldUrlScheme import fieldUrlScheme
from .httpStatus import httpStatus
from .isBool import isBool
from .isDict import isDict
from .isFloat import isFloat
from .isInt import isInt
from .isList import isList
from .isNone import isNone
from .isStr import isStr




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

import ast
import importlib
from .predicate import Predicate
from .helpers import dict2strkwargs, get_function_arguments
from .logger import Logger
from .globalConfig import GlobalConfig

_logger = Logger.get_logger(__name__)


class LeafPredicate(Predicate):
    """
    Leaf predicate representation
    """
    def __init__(self, predicate_func, node, flow, args=None):
        """
        :param predicate_func: predicate function
        :param node: node that predicate conforms to
        :param flow: flow to which predicate belongs to
        :param args: predicate arguments that should be used
        """
        self.node = node
        self.flow = flow

        self._func = predicate_func
        self._args = args if args is not None else {}
        self._func_args = get_function_arguments(self._func)

    def requires_message(self):
        """
        :return: True if predicate requires a message from a parent node
        """
        return 'message' in self._func_args

    def requires_node_args(self):
        """
        :return: True if predicate requires a node arguments
        """
        return 'node_args' in self._func_args

    def _check_parameters(self):
        """
        Check user defined predicate parameters against predicate parameters

        :raises: ValueError
        """
        func_args = get_function_arguments(self._func)
        user_args = self._args.keys()

        if 'message' in func_args:
            # message argument is implicit and does not need to be specified by user
            func_args.remove('message')

        if 'node_args' in func_args:
            # node_args are implicit as well
            func_args.remove('node_args')

        func_args = set(func_args)
        user_args = set(user_args)

        error = False

        for arg in func_args - user_args:
            _logger.error("Argument '%s' of predicate '%s' not specified in flow '%s'"
                          % (arg, self._func.__name__, self.flow.name))
            error = True

        for arg in user_args - func_args:
            _logger.error("Invalid argument '%s' for predicate '%s' in flow '%s'"
                          % (arg, self._func.__name__, self.flow.name))
            error = True

        if error:
            raise ValueError("Bad predicate arguments specified in flow '%s'" % self.flow.name)

    def _check_usage(self):
        """
        Check correct predicate usage

        :raises: ValueError
        """
        if self.requires_message() and self.node and self.node.is_flow():
            raise ValueError("Results of sub-flows cannot be used in predicates")
        if self.requires_message() and not self.node:
            raise ValueError("Cannot inspect results in starting edge in predicate '%s'" % self._func.__name__)
        if self.requires_message() and not self.node.storage:
            raise ValueError("Cannot use predicate that requires a message without storage '%s'" % self._func.__name__)

    def check(self):
        """
        Check whether predicate is correctly used

        :raises: ValueError
        """
        self._check_usage()
        self._check_parameters()

    def __str__(self):
        if self.requires_message():
            return "%s(db.get('%s'), %s)"\
                   % (self._func.__name__, self._task_str_name(), dict2strkwargs(self._args))
        else:
            # we hide node_args parameter
            return "%s(%s)" % (self._func.__name__, dict2strkwargs(self._args))

    def _task_str_name(self):
        # task_name can be None if we have starting edge
        if self.node is None:
            return 'None'
        else:
            return "%s" % self.node.name

    def ast(self):
        """
        :return: AST representation of predicate
        """
        # we could directly use db[task] in predicates, but predicates should not handle database errors,
        # so leave them on higher level (selinon) and index database before predicate is being called

        kwargs = []
        # we want to avoid querying to database if possible, if a predicate does not require message, do not ask for it
        if self.requires_message():
            # this can raise an exception if check was not run, since we are accessing storage that can be None
            kwargs.append(ast.keyword(arg='message',
                                      value=ast.Call(func=ast.Attribute(value=ast.Name(id='db', ctx=ast.Load()),
                                                                        attr='get', ctx=ast.Load()),
                                                     args = [ast.Str(s=self._task_str_name())],
                                                     keywords=[], starargs = None, kwargs = None)))
        if self.requires_node_args():
            kwargs.append(ast.keyword(arg='node_args', value=ast.Name(id='node_args', ctx=ast.Load())))

        kwargs.extend([ast.keyword(arg=k, value=ast.Str(s=v)) for k, v in self._args.items()])

        return ast.Call(func=ast.Name(id=self._func.__name__, ctx=ast.Load()),
                        args=[], starargs=None, kwargs=None, keywords=kwargs)

    def predicates_used(self):
        """
        :return: list of predicates that are used
        :rtype: List[Predicate]
        """
        return [self._func]

    @classmethod
    def create(cls, name, node, flow, args=None):
        """
        Create predicate

        :param name: predicate name
        :type name: str
        :param node: node to which predicate belongs
        :type node: Node
        :param flow: flow to which predicate belongs
        :type flow: Flow
        :param args: predicate arguments
        :return: an instantiated predicate
        :raises: ImportError
        """
        try:
            module = importlib.import_module(GlobalConfig.predicates_module)
            predicate = getattr(module, name)
        except ImportError:
            _logger.error("Cannot import predicate '{}'".format(name))
            raise
        return LeafPredicate(predicate, node, flow, args)



#!/usr/bin/env python3

from .buildinPredicate import AndPredicate, NotPredicate, OrPredicate, BuildinPredicate, NaryPredicate
from .buildinPredicate import UnaryPredicate
from .config import Config
from .edge import Edge
from .flow import Flow
from .leafPredicate import LeafPredicate
from .logger import Logger
from .node import Node
from .predicate import Predicate
from .system import System
from .task import Task
from .storage import Storage
from .version import selinonlib_version



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################


class Storage(object):
    """
    A storage representation
    """
    def __init__(self, name, import_path, configuration, class_name=None):
        """
        :param name: storage name
        :param import_path: storage import path
        :param configuration: storage configuration that will be passed
        :param class_name: storage class name
        """
        self.name = name
        self.import_path = import_path
        self.configuration = configuration
        self.class_name = class_name if class_name else name
        self.tasks = []

    def register_task(self, task):
        """
        Register a new that uses this storage

        :param task: task to be registered
        """
        self.tasks.append(task)

    @staticmethod
    def from_dict(d):
        """
        Construct storage instance from a dict

        :param d: dict that should be used to instantiate Storage
        :rtype: Storage
        """
        if 'name' not in d or not d['name']:
            raise KeyError('Storage name definition is mandatory')
        if 'import' not in d or not d['import']:
            raise KeyError('Storage import definition is mandatory')
        if 'configuration' not in d or not d['configuration']:
            raise KeyError('Storage configuration definition is mandatory')

        return Storage(d['name'], d['import'], d['configuration'], d.get('classname'))

    @property
    def var_name(self):
        """
        return: name of storage variable that will be used in config.py file
        """
        return "_storage_%s" % self.name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################
"""
A failure is basically a node in graph, that represents all permutations of possible cases of failures. Consider
having two failure conditions:

.. code-block:: yaml

  failures:
    - nodes:
          - Task1
          - Task2
          - Task3
      fallback:
          - FallbackTask1
    - nodes:
          - Task1
          - Task2
      fallback:
          - FallbackTask2

What we do here, we construct a graph of all possible permutations connected using edges that represent a node
that should be added to create a new permutation:

.. code-block:: yaml

   |-------------------------------
   |                              |
  T1           T2           T3    |
    \         /  \         / \    |
     \       /    \       /   \   |
      \     /      \     /     \  v
       T1,T2*       T2,T3      T1,T3
          \         /           /
           \       /           /
            T1,T2,T3*  <-------

For nodes ``T1,T2,T3`` and ``T1,T2`` we assign a fallback as configured. This graph is then serialized into the Python
configuration code. This way the dispatcher has O(N) time complexity when dealing with failures.

Note we are creating sparse tree - only for nodes that are listed in failures.

Note that we link failure nodes as allocated - we get a one way linked list of all failure nodes that helps us with
Python code generation.
"""

from functools import reduce


class FailureNode(object):
    """
    A representation of a failure node permutation
    """
    def __init__(self, flow, traversed, failure_link):
        """
        :param flow: flow to which the failure node conforms to
        :param traversed: traversed nodes - also permutation of nodes
        :param failure_link: link to next failure node in failures
        """
        self.next = {}
        self.flow = flow
        self.fallback = []
        self.traversed = traversed
        self.failure_link = failure_link

    def to(self, node_name):
        """
        Retrieve next permutation

        :param node_name: a name of the node for next permutation
        :rtype: FailureNode
        """
        return self.next[node_name]

    def has_to(self, node_name):
        """
        :param node_name:
        :return: True if there is a link to next permutation for node of name node_name
        """
        return node_name in self.next

    def add_to(self, node_name, failure):
        """
        Add failure for next permutation

        :param node_name: a node for next permutation
        :param failure: FailureNode that should be added
        """
        assert(node_name not in self.next)
        self.next[node_name] = failure

    @staticmethod
    def _add_fallback(failure_node, fallback):
        if len(failure_node.fallback) > 0:
            raise ValueError("Multiple definitions of a failure in flow '%s' with failure of %s"
                             % (failure_node.flow.name, failure_node.traversed))

        if not isinstance(fallback, list) and fallback is not True:
            fallback = [fallback]

        # additional checks are done by System
        failure_node.fallback = fallback

    @classmethod
    def construct(cls, flow, failures):
        """
        Construct failures from failures dictionary

        :param failures: failures dictionary
        :param flow: flow to which failures conform to
        :return: a link for linked list of failures and a dict of starting failures
        """
        last_allocated = None
        starting_failures = {}

        for failure in failures:
            used_starting_failures = {}

            for node in failure['nodes']:
                if node not in starting_failures:
                    f = FailureNode(flow, [node], last_allocated)
                    last_allocated = f
                    starting_failures[node] = f
                    used_starting_failures[node] = f
                else:
                    used_starting_failures[node] = starting_failures[node]

            current_nodes = list(used_starting_failures.values())
            next_nodes = []

            for _ in range(1, len(failure['nodes'])):  # for every permutation length

                for edge_node in failure['nodes']:  # edge_node is a node that can create a new permutation from a node

                    for current_node in current_nodes:

                        if edge_node not in current_node.traversed:

                            if not current_node.has_to(edge_node):
                                next_node = current_node.traversed + [edge_node]
                                f = FailureNode(flow, next_node, last_allocated)
                                last_allocated = f
                                current_node.add_to(edge_node, f)

                                for node in current_nodes:
                                    diff = set(node.traversed) ^ set(next_node)

                                    if len(diff) == 1:
                                        if not node.has_to(list(diff)[0]):
                                            node.add_to(list(diff)[0], f)

                                next_nodes.append(f)
                            else:
                                # keep for generating new permutations
                                next_nodes.append(current_node.to(edge_node))

                current_nodes = next_nodes
                next_nodes = []

            f = reduce(lambda x, y: x.to(y), failure['nodes'][1:], used_starting_failures[failure['nodes'][0]])
            cls._add_fallback(f, failure['fallback'])

        # we could make enumerable and avoid last_allocated (it would be cleaner), but let's stick with
        # this one for now
        return last_allocated, starting_failures



#!/bin/env python3

from functools import reduce


def argsFieldFloat(node_args, key):
    try:
        val = reduce(lambda m, k: m[k], key if isinstance(key, list) else [key], node_args)
        return isinstance(val, float)
    except:
        return False




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

import logging


class _LoggerWrapper(object):
    """
    Logger wrapper to be sure that logger is called with a correct logging level in the whole library after imports
    """
    def __init__(self, name):
        self._instance = None
        self._name = name

    def __getattr__(self, item):
        if item in ['_name', '_level', '_instance']:
            return super(_LoggerWrapper, self).__getattribute__(item)
        if not self._instance:
            logging.basicConfig(level=Logger.get_level())
            self._instance = logging.getLogger(self._name)
        return getattr(self._instance, item)


class Logger(object):
    """
    Library logger
    """
    _verbosity = 0

    @classmethod
    def set_verbosity(cls, verbosity):
        """
        :param verbosity: logger verbosity
        :type verbosity: int
        """
        cls._verbosity = verbosity

    @classmethod
    def get_level(cls):
        """
        :return: logging level for logging module
        """
        if not cls._verbosity:
            return logging.WARNING

        if cls._verbosity == 0:
            return logging.WARNING
        elif cls._verbosity == 1:
            return logging.INFO
        elif cls._verbosity > 1:
            return logging.DEBUG

    @classmethod
    def get_logger(cls, name):
        """
        Get logging logger

        :param name: logging logger name
        :rtype: _LoggerWrapper
        """
        return _LoggerWrapper(name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

from itertools import chain
from .failureNode import FailureNode


class Failures(object):
    """
    Node failures and fallback handling
    """
    def __init__(self, raw_definition, system, flow, last_allocated=None, starting_nodes=None):
        """
        :param raw_definition: raw definition of failures
        :param system: system context
        :param last_allocated: last allocated starting node for linked list
        :param starting_nodes: starting nodes for failures
        """
        self.waiting_nodes = []
        self.fallback_nodes = []

        for failure in raw_definition:
            waiting_nodes_entry = []
            for node_name in failure['nodes']:
                node = system.node_by_name(node_name, graceful=True)
                if not node:
                    raise KeyError("No such node with name '%s' in failure, flow '%s'" % (node_name, flow.name))
                waiting_nodes_entry.append(node)

            if isinstance(failure['fallback'], list):
                fallback_nodes_entry = []
                for node_name in failure['fallback']:
                    node = system.node_by_name(node_name, graceful=True)
                    if not node:
                        raise KeyError("No such node with name '%s' in failure fallback, flow '%s'"
                                       % (node_name, flow.name))
                    fallback_nodes_entry.append(node)
            elif isinstance(failure['fallback'], bool):
                fallback_nodes_entry = failure['fallback']
            else:
                raise ValueError("Unknown fallback definition in flow '%s', failure: %s" % (flow.name, failure))

            self.waiting_nodes.append(waiting_nodes_entry)
            self.fallback_nodes.append(fallback_nodes_entry)

        self.raw_definition = raw_definition
        self.last_allocated = last_allocated
        self.starting_nodes = starting_nodes

    def all_waiting_nodes(self):
        """
        :return: all nodes that for which there is defined a callback
        """
        return list(set(chain(*self.waiting_nodes)))

    def all_fallback_nodes(self):
        """
        :return: all nodes that are used as fallback nodes
        """
        # remove True/False flags
        nodes = []
        for fallback in self.fallback_nodes:
            for node in fallback:
                if not isinstance(node, bool):
                    nodes.append(node)
        return list(set(nodes))

    @staticmethod
    def construct(system, flow, failures_dict):
        """
        :param system: system context
        :param flow: a flow to which failures conform
        :param failures_dict: construct failures from failures dict
        :rtype: Failures
        """

        for failure in failures_dict:
            if 'nodes' not in failure or failure['nodes'] is None:
                raise ValueError("Failure should state nodes for state 'nodes' to fallback from in flow '%s'"
                                 % flow.name)

            if 'fallback' not in failure:
                raise ValueError("No fallback stated in failure in flow '%s'" % flow.name)

            if not isinstance(failure['nodes'], list):
                failure['nodes'] = [failure['nodes']]

            if not isinstance(failure['fallback'], list) and failure['fallback'] is not True:
                failure['fallback'] = [failure['fallback']]

            if failure['fallback'] is not True and len(failure['fallback']) == 1 and len(failure['nodes']) == 1 \
                    and failure['fallback'][0] == failure['nodes'][0]:
                raise ValueError("Detect cyclic fallback dependency in flow %s, failure on %s"
                                 % (flow.name, failure['nodes'][0]))

        last_allocated, starting_nodes = FailureNode.construct(flow, failures_dict)
        return Failures(failures_dict, system, flow, last_allocated, starting_nodes)

    @staticmethod
    def starting_nodes_name(flow_name):
        """
        A starting node name representation for generated Python config
        """
        return "_%s_failure_starting_nodes" % flow_name

    @staticmethod
    def failure_node_name(flow_name, failure_node):
        """
        A failure node name representation for generated Python config
        """
        return "_%s_fail_%s" % (flow_name, "_".join(failure_node.traversed))

    def fallback_nodes_names(self):
        """
        :return: names of nodes that are started by fallbacks in all failures
        """
        ret = []

        failure_node = self.last_allocated
        while failure_node:
            if isinstance(failure_node.fallback, list):
                ret.extend(failure_node.fallback)
            failure_node = failure_node.failure_link

        return ret

    def waiting_nodes_names(self):
        """
        :return: names of all nodes that we are expecting to fail for fallbacks
        """
        return list(self.starting_nodes.keys())

    def dump2stream(self, f, flow_name):
        """
        Dump failures to the Python config file for Dispatcher

        :param f: output stream to dump to
        :param flow_name: a name of a flow that failures belong to
        """
        fail_node = self.last_allocated

        while fail_node:
            next_dict = {}
            for k, v in fail_node.next.items():
                next_dict[k] = self.failure_node_name(flow_name, v)

            f.write("%s = {'next': " % self.failure_node_name(flow_name, fail_node))

            # print "next_dict"
            f.write('{')
            printed = False
            for k, v in next_dict.items():
                if printed:
                    f.write(", ")
                f.write("'%s': %s" % (k, v))
                printed = True
            f.write('}, ')

            # now list of nodes that should be started in case of failure (fallback)
            f.write("'fallback': %s}\n" % fail_node.fallback)
            fail_node = fail_node.failure_link

        f.write("\n%s = {" % self.starting_nodes_name(flow_name))

        printed = False
        for k, v in self.starting_nodes.items():
            if printed:
                f.write(",")
            f.write("\n    '%s': %s" % (k, self.failure_node_name(flow_name, v)))
            printed = True
        f.write("\n}\n\n")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ####################################################################
# Copyright (C) 2016  Fridolin Pokorny, fpokorny@redhat.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# ####################################################################

from .predicate import Predicate
from .buildinPredicate import AlwaysTruePredicate


class Edge(object):
    """
    Edge representation
    """
    def __init__(self, nodes_from, nodes_to, predicate, flow):
        """
        :param nodes_from: nodes from where edge starts
        :type nodes_from: List[Node]
        :param nodes_to: nodes where edge ends
        :type nodes_to: List[Node]
        :param predicate: predicate condition
        :type predicate: Predicate
        :param flow: flow to which edge belongs to
        :type flow: Flow
        """
        self.nodes_from = nodes_from
        self.nodes_to = nodes_to
        self.predicate = predicate
        self.flow = flow

    @staticmethod
    def from_dict(d, system, flow):
        """
        Construct edge from a dict

        :param d: a dictionary from which the system should be created
        :type d: dict
        :param system:
        :type system: System
        :param flow: flow to which edge belongs to
        :type flow: Flow
        :return:
        """
        if 'from' not in d:
            raise ValueError("Edge definition requires 'from' explicitly to be specified, use empty for starting edge")

        # we allow empty list for a starting edge
        if d['from']:
            from_names = d['from'] if isinstance(d['from'], list) else [d['from']]
            nodes_from = [system.node_by_name(n) for n in from_names]
        else:
            nodes_from = []

        if 'to' not in d or not d['to']:
            raise ValueError("Edge definition requires 'to' specified")

        to_names = d['to'] if isinstance(d['to'], list) else [d['to']]
        nodes_to = [system.node_by_name(n) for n in to_names]

        if 'condition' in d:
            predicate = Predicate.construct(d.get('condition'), nodes_from, flow)
        else:
            predicate = AlwaysTruePredicate(flow=flow)

        return Edge(nodes_from=nodes_from, nodes_to=nodes_to, predicate=predicate, flow=flow)



#!/usr/bin/python3

from setuptools import setup, find_packages

# Note: We are not distributing examples/ for now


def get_requirements():
    with open('requirements.txt') as fd:
        return fd.read().splitlines()


setup(
    name='selinonlib',
    version='0.1.0rc3',
    packages=find_packages(),
    scripts=['selinonlib-cli'],
    install_requires=get_requirements(),
    author='Fridolin Pokorny',
    author_email='fpokorny@redhat.com',
    maintainer='Fridolin Pokorny',
    maintainer_email='fpokorny@redhat.com',
    description='a simple tool to visualize, check and generate Python code from a YAML configuration file for Selinon'
                ' dispatcher for Celery',
    url='https://github.com/fridex/selinonlib',
    license='GPL',
    keywords='node task graph edge celery selinon yaml condition',
)
